unit ZDbcPooled;

interface

uses
  {$IFDEF MSWINDOWS}
  Windows,
  {$ELSE}
  {$ENDIF}
  Classes,
  SysUtils,
  zConnection;

type
  TZConnectionPool = class
  private
    FConnections     : array of TZConnection;
    FSlotsInUse      : TBits;
    FSlotsSize       : Integer;
    FCriticalSection : TRTLCriticalSection;
  public
    constructor Create;
    destructor Destroy; override;
    function Acquire(const Protocol, HostName, UserID, Password, DatabaseName : string) : TZConnection;
    procedure ReturnToPool(Connection: TZConnection);
  end;

implementation

{$I ZDbc.inc}
uses
  ZDbcIntfs,
  ZCompatibility;

{ TZConnectionPool }

constructor TZConnectionPool.Create;
begin
  inherited Create;
  FSlotsInUse := TBits.Create;
  InitializeCriticalSection(FCriticalSection);
  SetLength(FConnections, 10);
  FSlotsSize := 10;
  FSlotsInUse.Size := 10;
end;

destructor TZConnectionPool.Destroy;
var
  i : integer;
begin
  for i := 0 to Length(FConnections) - 1 do
  begin
    try
      FConnections[i].Free;
    except
      //
    end;
    FConnections[i] := nil;
  end;
  SetLength(FConnections, 0);
  FSlotsInUse.Free;
  DeleteCriticalSection(FCriticalSection);
  inherited Destroy;
end;


function TZConnectionPool.Acquire(const Protocol, HostName, UserID, Password, DatabaseName : string) : TZConnection;
var
  i          : Integer;
  Connection : TZConnection;

  function CanUseSlot(Connection : TZConnection) : boolean;
  begin
    Result := (Connection.Protocol = Protocol) and (Connection.HostName = HostName) and
      (Connection.User = UserID) and (Connection.Password = Password) and (Connection.Database = DatabaseName);
  end;

  function IsAlive(Connection : TZConnection) : boolean;
  begin
    result := Connection.Ping;
  end;

  function CreateConnection : TZConnection;
  begin
    Result := TZConnection.Create(nil);
    Result.HostName := HostName;
    Result.Protocol := Protocol;
    Result.User     := UserID;
    Result.Password := Password;
    Result.Database := DatabaseName;
  end;

begin

  EnterCriticalSection(FCriticalSection);
  try
    // Try to get an existing connection
    i := 0;
    while i < FSlotsSize do
    begin
      Connection := FConnections[i];
      if (Connection <> nil) and (not FSlotsInUse[i]) and CanUseSlot(Connection) then
      begin
        if IsAlive(Connection) then
        begin
          FSlotsInUse[i] := True;
          Break;
        end
        else
        begin
          try
            FConnections[i] := nil;
            Connection.Free;
          except
          end;
        end;
        Inc(i);
      end
      else
        Inc(i);
    end;

    if i = FSlotsSize then
    begin
      i := 0;
      while i < FSlotsSize do
      begin
        if (FConnections[i] = nil) and (not FSlotsInUse[i]) then
        begin
          FSlotsInUse[i] := True;
          Break;
        end;
        Inc(i);
      end;
    end;

    if i = FSlotsSize then
    begin
      SetLength(FConnections, Length(FConnections) + 10);
      FSlotsSize := FSlotsSize + 10;
      FSlotsInUse.Size := FSlotsSize;
      FSlotsInUse[i] := True;
    end;
  finally
    LeaveCriticalSection(FCriticalSection);
  end;

  if FConnections[i] = nil then
  begin
    Connection := CreateConnection;
    try
      Connection.Connect;
      FConnections[i] := Connection;
    except
      on E: Exception do
      begin
        EnterCriticalSection(FCriticalSection);
        try
          FSlotsInUse[i] := False;
          Connection.Free;
        finally
          LeaveCriticalSection(FCriticalSection);
          raise Exception.Create(ClassName + '.Acquire' + LineEnding + 'Error while trying to acquire a new connection' + LineEnding + LineEnding + E.Message);
        end;
      end;
    end;
  end;
  Result := TZConnection(FConnections[i]);
  Result.TransactIsolationLevel := tiReadCommitted;
end;

procedure TZConnectionPool.ReturnToPool(Connection: TZConnection);
var
  i : Integer;
begin
  EnterCriticalSection(FCriticalSection);
  try
    for i := 0 to Length(FConnections) - 1 do
    begin
      if FConnections[i] = Connection then
      begin
        FSlotsInUse[i] := False;
        Break;
      end;
    end;
  finally
    LeaveCriticalSection(FCriticalSection);
  end;
end;

end.

